import time

import h5py
import torch
from mpi4py import MPI

torch.set_default_dtype(torch.float32)


def dist_symmetric(x: torch.Tensor, y: torch.Tensor, comm: MPI.Comm = MPI.COMM_WORLD) -> torch.Tensor:
    """
    Calculate pairwise distances between all rows (samples, i.e., along axis 0) of two tensors x and y in parallel.

    The distance matrix is calculated tile-wise with ring communication between processes, each holding a piece of x
    and/or y.

    Parameters
    ----------
    x : torch.Tensor
        First 2d tensor (of shape m/p x f). m is the total number of samples in x, distributed over p processors.
        f is the number of features.
    y : torch.Tensor
        Second 2d tensor (of shape n/p x f). n is the total number of samples in x, distributed over p processors.
        The number of features f must be the same as for x.
    comm : MPI.Comm
        Communicator to use. Default is ``MPI.COMM_WORLD``.
    """
    # Check whether two input tensors are compatible.
    if len(x.shape) != len(y.shape) != 2:
        raise ValueError("Input tensors must be two-dimensional.")
    if x.shape[1] != y.shape[1]:
        raise ValueError(f"Input tensors must have the same number of features but {x.shape[1]} != {y.shape[1]}.")

    size, rank = comm.size, comm.rank  # Set up communication.

    if size == 1:  # Use torch functionality in non-parallel case.
        return torch.cdist(x, y)

    else:  # Parallel case
        # --- Setup and Matrix Initialization ---
        mp, f = x.shape  # Get number of samples in local chunk of x and number of features.
        np = y.shape[0]  # Get number of samples in local chunk of y.

        # Each process initializes a local matrix, `local_distances`, of shape `(mp, n)`, where `mp` is the local chunk
        # size of `x`, and `n` is the total number of samples in `y`. Each rank thus calculates the distance matrix
        # chunk of size `mp x n`, i.e., rank 0 has distances from its own local `x` to all other `y`'s.

        # Determine overall number of samples in y.
        n = comm.allreduce(np, op=MPI.SUM)
        print(f"Overall number of samples is {n}.")
        # Initialize rank-local chunk of mp x n distance matrix with zeros.
        local_distances = torch.zeros((mp, n))

        # --- Managing Chunks and Displacements ---
        # Determine where to put each result in the rank-local distance matrix chunk.
        # Determine number of samples (rows) in each rank-local y.
        y_counts = torch.tensor(comm.allgather(torch.numel(y) // f), dtype=torch.int)
        # Calculate corresponding displacements from counts to record the starting index of each chunk in y. Thus, each
        # process can identify where in the result matrix it should write the distances.
        y_displ = (0,) + tuple(torch.cumsum(y_counts, dim=0, dtype=torch.int)[:-1])
        # Extract actual result columns in distance matrix chunk for each rank.
        cols = (y_displ[rank], y_displ[rank + 1] if (rank + 1) != size else n)

        # --- Ring Communication Pattern ---
        # Calculate distances in a "ring" pattern. Each process calculates distances for its local x chunk against its
        # local y chunk (diagonal calculation). Then, through `size - 1` iterations, each process sends its y chunk to
        # the next process in the "ring" while receiving a new y chunk from the previous process. This continues until
        # each process has calculated distances between its x chunk and all chunks of y across all processes.
        x_ = x
        stationary = y

        # 0th iteration: Calculate diagonal of global distance matrix.
        # Each process calculates distances for its local x chunk against its local y chunk.
        print(f"Rank [{rank}/{size}]: Calculate diagonal blocks...")
        d_ij = torch.cdist(x_, stationary)
        local_distances[:, cols[0]: cols[1]] = d_ij

        print(f"Rank [{rank}/{size}]: Start tile-wise ring communication...")
        # Remaining `(size+1) // 2` iterations: Send rank-local part of y to next process in circular fashion.
        # We can perform less iterations due to the symmetric nature of the metric.
        for iter_idx in range(1, (size + 2) // 2):
            print(f"Rank [{rank}/{size}]: Starting iteration {iter_idx}")
            # Since in the final iteration (with and even size) we only need to send from the first half of the
            # processes, we check if this is the final iteration.
            is_final_iteration = iter_idx == (size + 1) // 2
            receiver = (rank + iter_idx) % size  # Determine receiving process.
            sender = (rank - iter_idx) % size  # Determine sending process.
            # Determine columns of rank-local distance matrix chunk to write result to.
            col1 = y_displ[sender]
            col2 = y_displ[sender + 1] if sender != size - 1 else n
            columns = (col1, col2)

            # Since the final iteration is different, we handle the communication differently.
            if is_final_iteration:
                # The second half of the processes only need to receive.
                if rank >= size // 2:
                    stat = MPI.Status()
                    # Probe for incoming message containing the next chunk of y to consider.
                    comm.Probe(source=sender, tag=iter_idx, status=stat)
                    # Determine number of samples to receive (= overall number of floats to receive / number of features).
                    count = int(stat.Get_count(MPI.FLOAT) / f)
                    # Initialize tensor for incoming chunk of y with zeros.
                    moving = torch.zeros((count, f))
                    comm.Recv(moving, source=sender, tag=iter_idx)
                    # Calculate distances between stationary chunk of x and currently considered, moving chunk of y.
                    # Write result at correct position in distance matrix.
                    d_ij = torch.cdist(x_, moving)
                    local_distances[:, columns[0]: columns[1]] = d_ij
                # The first half of the processes only need to send.
                if rank < size // 2:
                    comm.Send(stationary, dest=receiver, tag=iter_idx)
            # Standard communication for all but the final iteration.
            else:
                # All but first `iter_idx` processes are first receiving, then sending.
                if (rank // iter_idx) != 0:
                    stat = MPI.Status()
                    # Probe for incoming message containing the next chunk of y to consider.
                    comm.Probe(source=sender, tag=iter_idx, status=stat)
                    # Determine number of samples to receive (= overall number of floats to receive / number of features).
                    count = int(stat.Get_count(MPI.FLOAT) / f)
                    # Initialize tensor for incoming chunk of y with zeros.
                    moving = torch.zeros((count, f))
                    comm.Recv(moving, source=sender, tag=iter_idx)
                # Send rank-local chunk of y to next process.
                comm.Send(stationary, dest=receiver, tag=iter_idx)
                # First `iter_idx` processes can now receive after sending.
                if (rank // iter_idx) == 0:
                    stat = MPI.Status()
                    comm.Probe(source=sender, tag=iter_idx, status=stat)
                    count = int(stat.Get_count(MPI.FLOAT) / f)
                    moving = torch.zeros((count, f))
                    comm.Recv(moving, source=sender, tag=iter_idx)
                # Calculate distances between stationary chunk of x and currently considered, moving chunk of y.
                # Write result at correct position in distance matrix.
                d_ij = torch.cdist(x_, moving)
                local_distances[:, columns[0]: columns[1]] = d_ij

            # Step 2.3 - Only for symmetric metrics:
            # We want to send the calculated distances back to process they came from since with symmetric metrics
            # d(i,j) = d(j,i) and we can save calculation.
            # First, we determine the columns of the distance matrix that the receiver will need to write.
            symmetric_col1 = y_displ[receiver]
            symmetric_col2 = y_displ[receiver + 1] if receiver != size - 1 else n
            symmetric_columns = (symmetric_col1, symmetric_col2)

            # Now, we send the calculated distances back to the symmetric tiles.
            # We need to calculate the chunk size, therefore subtracting the smaller column index from the larger
            # column index gives us the number of samples.
            symmetric = torch.zeros(symmetric_columns[1] - symmetric_columns[0], (cols[1] - cols[0]))

            # Communication for the final iteration.
            if is_final_iteration:
                # Now the first half of the processes are receiving the symmetrical values back.
                if rank < size // 2:
                    comm.Recv(symmetric, source=receiver, tag=iter_idx)
                    # Save into the local distance matrix.
                    local_distances[:, symmetric_columns[0]: symmetric_columns[1]] = symmetric.transpose(0, 1)
                # The second half that performed the calculation send the results back.
                if rank >= size // 2:
                    comm.Send(d_ij, dest=sender, tag=iter_idx)
                    # Save the local distance matrix, nothing is received.
                    local_distances[:, symmetric_columns[0]: symmetric_columns[1]] = d_ij
            # Standard communication handling.
            else:
                # As before, all but first `iter_idx` processes are first receiving, then sending.
                if (rank // iter_idx) != 0:
                    comm.Recv(symmetric, source=receiver, tag=iter_idx)
                # Then send.
                comm.Send(d_ij, dest=sender, tag=iter_idx)
                # Now the first `iter_idx` process can also receive after sending.
                if (rank // iter_idx) == 0:
                    comm.Recv(symmetric, source=receiver, tag=iter_idx)
                # Save into the local distance matrix.
                local_distances[:, symmetric_columns[0]: symmetric_columns[1]] = symmetric.transpose(0, 1)
        print(f"Rank [{rank}/{size}]: [DONE]")

        return local_distances


if __name__ == "__main__":
    comm = MPI.COMM_WORLD
    rank, size = comm.rank, comm.size

    data_path = "./SUSY_50k.h5"
    dataset = "data"

    if rank == 0:
        print(
            "######################\n"
            "# Pairwise distances #\n"
            "######################\n"
            f"COMM_WORLD size is {size}.\n"
            f"Loading data... {data_path}[{dataset}]"
        )

    # Parallel data loader for SUSY data.
    with h5py.File(data_path, "r") as handle:
        chunk = int(handle[dataset].shape[0] / size)
        if rank == size - 1:
            data = torch.FloatTensor(handle[dataset][rank * chunk:])
        else:
            data = torch.FloatTensor(handle[dataset][rank * chunk:(rank + 1) * chunk])

    print(f"\t[OK]\nRank [{rank}/{size}]: Local data chunk has shape {list(data.shape)}...")

    if rank == 0:
        print("Start distance calculations...")
    # Calculate distances of all SUSY samples w.r.t. each other and measure runtime.
    start = time.perf_counter()
    distances = dist_symmetric(data, data, comm)
    local_runtime = time.perf_counter() - start
    # Calculate process-averaged runtime.
    average_runtime = comm.allreduce(local_runtime, op=MPI.SUM) / size
    print(f"Rank [{rank}/{size}]: Local distance matrix has shape {list(distances.shape)}.")
    if rank == 0:
        print(f"Process-averaged run time:\t{average_runtime} s")
